require 'json'
require 'uu/os/lang/json_patch'
require 'uu/os/lang/uniform_hash'
require 'uu/os/lang/hash_accessor'
require 'uu/os/util/name_converter'
require 'uu/os/util/url_converter'
require 'uu/os/server/parameter_parser'
require 'uu/os/server/abstract_servlet'
require 'uu/os/ipc/ipc_error'
require 'uu/os/ipc/error_kind'
require 'uu/os/ipc/error_codes'
require 'uu/os/application/application_error'
require 'uu/os/application/error_codes'
require 'uu/os/application/controller_loader'
require 'uu/os/server/immediate_response'
require 'uu/os/application/log_helper'
require 'uu/os/application/app_config_loader'

module UU
  module OS
    module Application

      # Ancestor for all controllers which are agnostic to particular server
      # (in terms of technology) on which is controller deployed.
      class AbstractController < UU::OS::Server::AbstractServlet
        include UU::OS::Util::NameConverter
        include UU::OS::Application::LogHelper

        # Main method for processing requests. By default it:
        #
        # * Asserts invocation method using {#assert_invocation_method} method.
        # * Asserts request content type using {#assert_content_type} method.
        # * Obtains main object using {#get_main_object_uri} method.
        # * Obtains parameters using {#get_parameters} method.
        # * Determine action method using {#get_action_method} method.
        # * Does necessary initializations using {#prepare_execution} method.
        # * Prepares context using {#get_context} method.
        # * Invokes action method.
        # * Converts action method result into expected response object.
        # * Does cleanup using {#finalize_execution} method.
        #
        # @param [UU::OS::HTTP::HTTPRequest] request Incoming request.
        # @return [Object] Response object (same value as returned by {#response},
        #    {#file_download_response} or {#streamed_response} methods).
        # @note Not intended to be used from controller implementation but for
        #    optional overriding of default behavior by custom implementation!
        # @private
        def service(request) # :nodoc:
          begin
            ctx = get_context(request)
            assert_invocation_method(ctx)
            assert_content_type(ctx)
            prepare_execution(ctx)
            action_method = get_action_method(ctx)
            result = self.send(action_method, ctx)
          ensure
            begin
              finalize_execution(ctx)
            rescue Exception => ef
              log_error(request.logger, 'Finalization of request failed.', ef)
            end
          end
          if result.nil?
            return response(204, {}, result)
          elsif result.kind_of?(Array) && (result.size == 3) && result[0].kind_of?(Fixnum) && result[1].kind_of?(Hash) && result[2].respond_to?(:each)
            return result
          elsif result.respond_to?(:read)
            return response(200, {content_type: 'application/json'}, result)
          else
            return response(200, {content_type: 'application/json'}, result.to_json)
          end
        end

        # @private
        def get_context(request)
          ctx = request.env['uu.context']
          if !ctx.kind_of?(UU::OS::Application::UseCaseContext)
            raise UU::OS::Application::ApplicationError.new(UU::OS::IPC::ErrorKind::SERVER, UU::OS::Application::ErrorCodes::UNEXPECTED_PROBLEM, "Application context is not available.")
          end
          return ctx
        end

        # Checks if invocation method (POST, GET, etc.) matches controller type.
        # @param [UU::OS::HTTP::HTTPRequest] request Incoming request.
        # @return [void]
        # @raise [UU::OS::IPC::IPCError] If invocation method does not match.
        # @note Not intended to be used from controller implementation but for
        #    optional overriding of default behavior by custom implementation!
        # @private
        def assert_invocation_method(ctx) # :nodoc:
          # No assertion for abstract controller
        end

        # Checks if incoming request body has valid content type.
        # By default +application/json+ or +multipart/form-data+
        # is expected.
        # @param [UU::OS::HTTP::HTTPRequest] request Incoming request.
        # @return [void]
        # @raise [UU::OS::IPC::IPCError] In case content type does not match
        #    expected one.
        # @note Not intended to be used from controller implementation but for
        #    optional overriding of default behavior by custom implementation!
        # @private
        def assert_content_type(ctx) # :nodoc:
          # No assertion for abstract controller
        end

        # Returns name of action method to be invoked on controller.
        # by default action method is constructed from last request
        # path element converted to +snake_case+ with +on_+ prefix.
        # @param [UU::OS::HTTP::HTTPRequest] request Incoming request.
        # @param [String] uu_uri URI of main object.
        # @param [Hash] parameters Parsed and deserialized parameters.
        # @return [Symbol, String] Name of action method to be invoked.
        # @note Not intended to be used from controller implementation but for
        #    optional overriding of default behavior by custom implementation!
        # @private
        def get_action_method(ctx) # :nodoc:
          action = ctx.request.path.split('/')[-1]
          action_method = 'on_' << to_snake_case(action)
          return action_method.to_sym
        end

        # Method is invoked before execution of control method. It should
        # be used for necessary initializations of environment. In case
        # method is overridden, new implementation should always use +super+
        # call to ensure all initializations are done.
        # @param [UU::OS::CON::HTTPRequest] request Incoming request.
        # @param [String] uu_uri URI of main object.
        # @param [Hash] parameters parsed parameters.
        # @return [void]
        # @raise [Exception] In case initialization of environment fails.
        # @note Not intended to be used from controller implementation but for
        #    optional overriding of default behavior by custom implementation!
        # @private
        def prepare_execution(ctx) # :nodoc:

        end

        # Method is always invoked after execution of control method (even if
        # method raised exception). It should be used to cleanup environment
        # (cleanup everything what was initialized in {#prepare_execution}
        # method). In case method is overridden, new implementation should always
        # use +super+ call to ensure cleanup is complete. This method must not
        # raise any exception.
        # @param [UU::OS::CON::HTTPRequest] request Incoming request.
        # @param [String, nil] uu_uri URI of main object or +nil+ if
        #    method is invoked due to error before request parameters are parsed.
        # @param [Hash, nil] parameters Parsed parameters or +nil+ if method is
        #    invoked before request parameters are parsed.
        # @return [void]
        # @note Not intended to be used from controller implementation but for
        #    optional overriding of default behavior by custom implementation!
        # @private
        def finalize_execution(ctx) # :nodoc:

        end

        private

        # Transforms invocation of unimplemented controller action method
        # into valid exception response.
        # @private
        def method_missing(method, *args, &block) # :nodoc:
          if method =~ /^on_/ && args.size == 1
            raise UU::OS::Application::ApplicationError.new(UU::OS::IPC::ErrorKind::CLIENT, UU::OS::Application::ErrorCodes::NOT_IMPLEMENTED,
            	"Functionality addressed by #{args[0].request.url} is not implemented.")
          else
            super(method, *args)
          end
        end

      end
    end
  end
end
